class AiogramTonconnectException(Exception):
    """
    Base exception class for Aiogram AiogramTonConnect custom exceptions.
    """


class LanguageCodeNotSupported(AiogramTonconnectException):
    """
    Exception raised when an unsupported language code is encountered.
    """


# List of error messages related to editing messages
MESSAGE_EDIT_ERRORS = [
    "no text in the message",
    "message can't be edited",
    "message is not modified",
    "message to edit not found",
]

# List of error messages related to deleting messages
MESSAGE_DELETE_ERRORS = [
    "message can't be deleted",
    "message to delete not found",
]



from typing import Callable, Dict, Any, Awaitable, Optional, Type, List
from aiogram import BaseMiddleware
from aiogram.fsm.context import FSMContext
from aiogram.types import TelegramObject, User
from redis.asyncio import Redis

from .manager import ATCManager
from .tonconnect import AiogramTonConnect
from .tonconnect.storage import SessionStorage
from .tonconnect.models import ATCUser, AppWallet, AccountWallet
from .utils.keyboards import InlineKeyboardBase, InlineKeyboard
from .utils.qrcode import QRCodeProviderBase, QRCodeProvider
from .utils.texts import TextMessageBase, TextMessage


class AiogramTonConnectMiddleware(BaseMiddleware):
    """
    Aiogram middleware for AiogramTonConnect integration.

    :param redis: Redis instance for storage.
    :param manifest_url: URL to the AiogramTonConnect manifest.
    :param exclude_wallets: List of wallet names to exclude.
    :param qrcode_provider: QRCodeProviderBase instance for generating QR codes.
    :param text_message: TextMessageBase class for managing text messages.
    :param inline_keyboard: InlineKeyboardBase class for managing inline keyboards.
    """

    def __init__(
            self,
            redis: Redis,
            manifest_url: str,
            exclude_wallets: List[str] = None,
            qrcode_provider: Optional[QRCodeProviderBase] = None,
            text_message: Optional[Type[TextMessageBase]] = None,
            inline_keyboard: Optional[Type[InlineKeyboardBase]] = None,
    ) -> None:
        self.redis = redis
        self.manifest_url = manifest_url
        self.exclude_wallets = exclude_wallets

        if not qrcode_provider:
            qrcode_provider = QRCodeProvider()
        self.qrcode_provider = qrcode_provider

        if not text_message:
            text_message = TextMessage
        self.text_message = text_message

        if not inline_keyboard:
            inline_keyboard = InlineKeyboard
        self.inline_keyboard = inline_keyboard

    async def __call__(
            self,
            handler: Callable[[TelegramObject, Dict[str, Any]], Awaitable[Any]],
            event: TelegramObject,
            data: Dict[str, Any],
    ) -> Any:
        """
        Execute middleware

        :param handler: Wrapped handler in middlewares chain
        :param event: Incoming event (Subclass of :class:`aiogram.types.base.TelegramObject`)
        :param data: Contextual data. Will be mapped to handler arguments
        :return: :class:`Any`
        """
        state: FSMContext = data.get("state")
        state_data = await state.get_data()

        account_wallet = state_data.get("account_wallet", None)
        account_wallet = AccountWallet(**account_wallet) if account_wallet else None
        data["account_wallet"] = account_wallet

        app_wallet = state_data.get("app_wallet", None)
        app_wallet = AppWallet.from_dict(app_wallet) if app_wallet else None
        data["app_wallet"] = app_wallet

        user: User = data.get("event_from_user", None)
        language_code = state_data.get("language_code", None)
        language_code = language_code if language_code else user.language_code
        last_transaction_boc = state_data.get("last_transaction_boc", None)

        atc_user = ATCUser(
            id=user.id,
            language_code=language_code,
            last_transaction_boc=last_transaction_boc,
            app_wallet=app_wallet,
            account_wallet=account_wallet,
        )
        data["atc_user"] = atc_user

        tonconnect = AiogramTonConnect(
            storage=SessionStorage(self.redis, atc_user.id),
            manifest_url=self.manifest_url,
            exclude_wallets=self.exclude_wallets
        )
        await tonconnect.restore_connection()
        atc_manager = ATCManager(
            redis=self.redis,
            tonconnect=tonconnect,
            qrcode_provider=self.qrcode_provider,
            text_message=self.text_message(atc_user.language_code),
            inline_keyboard=self.inline_keyboard(atc_user.language_code),
            user=atc_user,
            data=data,
        )
        data["atc_manager"] = atc_manager

        return await handler(event, data)



from aiogram.fsm.state import StatesGroup, State


class TcState(StatesGroup):
    """
    States group for managing different states in the AiogramTonConnect bot.

    States:
    - connect_wallet: State representing the process of connecting a wallet.
    - connect_wallet_timeout: State representing the timeout during the connection of a wallet.
    - send_transaction: State representing the process of sending a transaction.
    - send_transaction_timeout: State representing the timeout during the sending of a transaction.
    - send_transaction_rejected: State representing the rejection of a transaction.
    """
    connect_wallet = State()
    connect_wallet_timeout = State()
    send_transaction = State()
    send_transaction_timeout = State()
    send_transaction_rejected = State()



from dataclasses import dataclass, asdict
from typing import Awaitable, Callable, Dict, List, Optional


@dataclass
class Base:
    """
    Base data class providing a method to convert the instance to a dictionary.
    """

    def to_dict(self) -> Dict:
        """
        Convert the instance to a dictionary.

        :return: Dictionary representation of the instance.
        """
        return asdict(self)

    @classmethod
    def from_dict(cls, data: Dict) -> "Base":
        """
        Create an instance of the Base class from a dictionary.

        :param data: A dictionary containing data to initialize the instance.
        :return: An instance of the class.
        """
        data = {k: v for k, v in data.items() if k in cls.__annotations__}
        return cls(**data)  # noqa


@dataclass
class AppWallet(Base):
    """
    Data class representing a AiogramTonConnect application wallet.

    :param app_name: Name of the application.
    :param name: Name of the wallet.
    :param image: URL of the wallet image.
    :param bridge: List of dictionaries representing the bridge.
    :param bridge_url: Optional URL of the bridge.
    :param platforms: Optional list of supported platforms.
    :param tondns: Optional TON DNS information.
    :param about_url: Optional URL providing information about the wallet.
    :param universal_url: Optional universal URL associated with the wallet.
    """

    app_name: str
    name: str
    image: str
    bridge: List[Dict]
    bridge_url: Optional[str] = None
    platforms: List[str] = None
    tondns: Optional[str] = None
    about_url: Optional[str] = None
    universal_url: Optional[str] = None


@dataclass
class AccountWallet(Base):
    """
    Data class representing a AiogramTonConnect account wallet.

    :param address: Optional wallet address.
    :param state_init: Optional wallet state initialization.
    :param public_key: Optional public key associated with the wallet.
    :param chain: Optional chain information.
    """

    address: Optional[str] = None
    state_init: Optional[str] = None
    public_key: Optional[str] = None
    chain: Optional[int] = None


@dataclass
class ATCUser(Base):
    """
    Data class representing a AiogramTonConnect user.

    :param id: User ID.
    :param language_code: Language code associated with the user.
    :param app_wallet: Optional AiogramTonConnect application wallet.
    :param account_wallet: Optional AiogramTonConnect account wallet.
    :param last_transaction_boc: Optional BOC of the last transaction.
    """

    id: int
    language_code: str
    app_wallet: Optional[AppWallet] = None
    account_wallet: Optional[AccountWallet] = None
    last_transaction_boc: Optional[str] = None


@dataclass
class TransactionMessage(Base):
    """
    Data class representing a AiogramTonConnect transaction message.

    :param address: Wallet address associated with the message.
    :param payload: Payload of the message.
    :param amount: Amount associated with the message.
    """

    address: str
    payload: str
    amount: str


@dataclass
class Transaction(Base):
    """
    Data class representing a AiogramTonConnect transaction.

    :param valid_until: Validity duration of the transaction.
    :param messages: List of transaction messages.
    """

    valid_until: int
    messages: List[TransactionMessage]


@dataclass
class ConnectWalletCallbacks(Base):
    """
    Data class representing callbacks for connecting a wallet.

    :param before_callback: Callable function to be executed before connecting the wallet.
    :param after_callback: Callable function to be executed after connecting the wallet.
    """

    before_callback: Callable[[...], Awaitable]
    after_callback: Callable[[...], Awaitable]


@dataclass
class SendTransactionCallbacks(Base):
    """
    Data class representing callbacks for sending a transaction.

    :param before_callback: Callable function to be executed before sending the transaction.
    :param after_callback: Callable function to be executed after sending the transaction.
    """

    before_callback: Callable[[...], Awaitable]
    after_callback: Callable[[...], Awaitable]



from typing import List, Optional

from pytonconnect import TonConnect as BaseTonConnect
from pytonconnect.storage import IStorage

from .wallet.manager import WalletManager
from .models import AppWallet


class AiogramTonConnect(BaseTonConnect):
    """
    AiogramTonConnect class that extends the base AiogramTonConnect functionality.

    :param storage: An instance of IStorage for data storage.
    :param manifest_url: URL to the manifest file.
    :param exclude_wallets: Optional list of wallet names to exclude.
    :param args: Additional positional arguments.
    :param kwargs: Additional keyword arguments.
    """

    def __init__(
            self,
            storage: IStorage,
            manifest_url: str,
            exclude_wallets: Optional[List[str]] = None,
            *args,
            **kwargs,
    ) -> None:
        """
        Initialize AiogramTonConnect.

        :param storage: An instance of IStorage for data storage.
        :param manifest_url: URL to the manifest file.
        :param exclude_wallets: Optional list of wallet names to exclude.
        :param args: Additional positional arguments.
        :param kwargs: Additional keyword arguments.
        """
        kwargs["storage"] = storage
        kwargs["manifest_url"] = manifest_url
        super().__init__(*args, **kwargs)

        self.wallet_manager = WalletManager(exclude_wallets=exclude_wallets)

    async def get_wallets(self=None) -> List[AppWallet]:
        """
        Get the list of AiogramTonConnect wallets.

        :return: List of AiogramTonConnect application wallet instances.
        """
        return await self.wallet_manager.get_wallets()



import time
from base64 import urlsafe_b64encode
from typing import Optional, Union

from pytoniq_core import begin_cell

from .models import Transaction, TransactionMessage

__all__ = [
    "TONTransferTransaction",
    "NFTTransferTransaction",
]


class TONTransferTransaction(Transaction):
    """
    Create a TON (Telegram Open Network) transfer transaction.

    :param address: The address to which the transfer is made.
    :param amount: The amount to be transferred.
    :param comment: An optional comment for the transaction.
    """

    def __init__(
            self,
            address: str,
            amount: Union[int, float],
            comment: Optional[str] = "",
    ) -> None:
        payload = urlsafe_b64encode(
            begin_cell()
            .store_uint(0, 32)
            .store_string(comment)
            .end_cell()
            .to_boc()
        ).decode()
        super().__init__(
            messages=[
                TransactionMessage(
                    address=address,
                    payload=payload,
                    amount=str(int(amount * 10 ** 9)),
                ),
            ],
            valid_until=int(time.time() + 300),
        )


class JettonTransferTransaction(Transaction):
    """
    Create a Jetton transfer transaction.

    :param jetton_wallet_address: The address of the Jetton wallet.
    :param recipient_address: The recipient's address.
    :param jetton_amount: The amount of Jetton to transfer.
    :param jetton_decimal: The number of decimal places in the Jetton amount (default is 9).
    :param response_address: The address for the response (default is the recipient's address).
    """

    def __init__(
            self,
            jetton_wallet_address: str,
            recipient_address: str,
            jetton_amount: int,
            jetton_decimal: int = 9,
            response_address: str = None,
    ) -> None:
        payload = urlsafe_b64encode(
            begin_cell()
            .store_uint(0xf8a7ea5, 32)
            .store_uint(0, 64)
            .store_coins(int(jetton_amount * 10 ** jetton_decimal))
            .store_address(recipient_address)
            .store_address(response_address or recipient_address)
            .store_uint(0, 1)
            .store_coins(1)
            .store_uint(0, 1)
            .end_cell()
            .to_boc()
        ).decode()
        super().__init__(
            messages=[
                TransactionMessage(
                    address=jetton_wallet_address,
                    payload=payload,
                    amount=str(int(0.07 * 10 ** 9)),
                ),
            ],
            valid_until=int(time.time() + 300),
        )


class NFTTransferTransaction(Transaction):
    """
    Create an NFT (Non-Fungible Token) transfer transaction.

    :param nft_address: The address of the NFT.
    :param recipient_address: The recipient's address.
    :param response_address: The address for the response (default is the recipient's address).
    :param transfer_fee: The transfer fee amount (default is 0.05 TON).
    """

    def __init__(
            self,
            nft_address: str,
            recipient_address: str,
            response_address: str = None,
            transfer_fee: Union[int, float] = 0.05,
    ) -> None:
        payload = urlsafe_b64encode(
            begin_cell()
            .store_uint(0x5fcc3d14, 32)
            .store_uint(0, 64)
            .store_address(recipient_address)
            .store_address(response_address or recipient_address)
            .store_uint(0, 1)
            .store_coins(1)
            .store_uint(0, 1)
            .end_cell()
            .to_boc()
        ).decode()
        super().__init__(
            messages=[
                TransactionMessage(
                    address=nft_address,
                    payload=payload,
                    amount=str(int(transfer_fee * 10 ** 9)),
                ),
            ],
            valid_until=int(time.time() + 300),
        )



import pickle
from typing import Union

from redis.asyncio import Redis

from ..models import ConnectWalletCallbacks, SendTransactionCallbacks


class CallbackStorageBase:
    """
    Base class for callback storage.

    This class defines the common methods for handling callback storage.
    """

    def __init__(
            self,
            redis: Redis,
            user_id: int,
            collection: str,
    ) -> None:
        """
        Initialize the CallbackStorageBase.

        :param redis: Redis instance for asynchronous operations.
        :param user_id: User ID associated with the storage.
        :param collection: Collection name for storing callbacks.
        """
        self.redis = redis
        self.user_id = user_id
        self.collection = collection

    async def get(self) -> Union[ConnectWalletCallbacks, SendTransactionCallbacks]:
        """
        Get stored callbacks.

        :return: Callbacks object.
        """
        raise NotImplementedError

    async def add(self, callbacks: Union[ConnectWalletCallbacks, SendTransactionCallbacks]) -> None:
        """
        Add callbacks to the storage.

        :param callbacks: Callbacks object to be added.
        """
        raise NotImplementedError

    async def remove(self) -> None:
        """
        Remove stored callbacks.
        """
        raise NotImplementedError


class ConnectWalletCallbackStorage(CallbackStorageBase):

    def __init__(
            self,
            redis: Redis,
            user_id: int,
            collection: str = "ConnectWalletCallbacks",
    ) -> None:
        super().__init__(redis, user_id, collection)

    async def get(self) -> ConnectWalletCallbacks:
        async with self.redis.client() as client:
            serialized_value = await client.hget(self.collection, key=str(self.user_id))
            return ConnectWalletCallbacks(**pickle.loads(serialized_value)) if serialized_value else None

    async def add(self, connect_wallet_callbacks: ConnectWalletCallbacks) -> None:
        async with self.redis.client() as client:
            serialized_value = pickle.dumps(connect_wallet_callbacks.to_dict())
            await client.hset(self.collection, key=str(self.user_id), value=serialized_value)  # noqa

    async def remove(self) -> None:
        async with self.redis.client() as client:
            await client.hdel(self.collection, [str(self.user_id)])


class SendTransactionCallbackStorage(CallbackStorageBase):

    def __init__(
            self,
            redis: Redis,
            user_id: int,
            collection: str = "SendTransactionsCallbacks",
    ) -> None:
        super().__init__(redis, user_id, collection)

    async def get(self) -> SendTransactionCallbacks:
        async with self.redis.client() as client:
            serialized_value = await client.hget(self.collection, key=str(self.user_id))
            return SendTransactionCallbacks(**pickle.loads(serialized_value)) if serialized_value else None

    async def add(self, callbacks: SendTransactionCallbacks) -> None:
        async with self.redis.client() as client:
            serialized_value = pickle.dumps(callbacks.to_dict())
            await client.hset(self.collection, key=str(self.user_id), value=serialized_value)  # noqa

    async def remove(self) -> None:
        async with self.redis.client() as client:
            await client.hdel(self.collection, [str(self.user_id)])



from __future__ import annotations

import asyncio
from contextlib import suppress
from typing import Dict, Any, Union, Optional

from aiogram import Bot
from aiogram.exceptions import TelegramBadRequest
from aiogram.fsm.context import FSMContext
from aiogram.types import Message, InlineKeyboardMarkup
from aiogram.utils.markdown import hide_link
from pytonconnect.exceptions import (
    UserRejectsError,
    WalletNotConnectedError,
)
from redis.asyncio import Redis

from .tonconnect import AiogramTonConnect
from .tonconnect.storage import (
    ConnectWalletCallbackStorage,
    SendTransactionCallbackStorage,
    TaskStorage,
)
from .tonconnect.models import (
    AccountWallet,
    AppWallet,
    ConnectWalletCallbacks,
    SendTransactionCallbacks,
    Transaction,
    ATCUser,
)
from .utils.exceptions import (
    LanguageCodeNotSupported,
    MESSAGE_DELETE_ERRORS,
    MESSAGE_EDIT_ERRORS,
)
from .utils.keyboards import InlineKeyboardBase
from .utils.qrcode import QRCodeProvider
from .utils.states import TcState
from .utils.texts import TextMessageBase


class ATCManager:
    """
    Manager class for AiogramTonConnect integration.

    :param user: The AiogramTonConnect user.
    :param redis: Redis instance for storage.
    :param tonconnect: AiogramTonConnect instance.
    :param qrcode_provider: QRCodeProvider instance for generating QR codes.
    :param text_message: TextMessageBase class for managing text messages.
    :param inline_keyboard: InlineKeyboardBase class for managing inline keyboards.
    :param data: Additional data.
    :param emoji: Emoji string. Defaults to "💎".
    """

    def __init__(
            self,
            user: ATCUser,
            redis: Redis,
            tonconnect: AiogramTonConnect,
            qrcode_provider: QRCodeProvider,
            text_message: TextMessageBase,
            inline_keyboard: InlineKeyboardBase,
            data: Dict[str, Any],
            emoji: str = "💎",
    ) -> None:
        self.user: ATCUser = user
        self.redis: Redis = redis
        self.tonconnect: AiogramTonConnect = tonconnect

        self.__qrcode_provider: QRCodeProvider = qrcode_provider
        self.__text_message: TextMessageBase = text_message
        self.__inline_keyboard: InlineKeyboardBase = inline_keyboard

        self.bot: Bot = data.get("bot")
        self.state: FSMContext = data.get("state")

        self.__data: Dict[str, Any] = data
        self.__emoji: str = emoji

        self.task_storage = TaskStorage(user.id)
        self.connect_wallet_callbacks_storage = ConnectWalletCallbackStorage(redis, user.id)
        self.send_transaction_callbacks_storage = SendTransactionCallbackStorage(redis, user.id)

    @property
    def middleware_data(self) -> Dict[str, Any]:
        """
        Get middleware data.
        """
        return self.__data

    async def update_interfaces_language(self, language_code: str) -> None:
        """
        Update interfaces language.

        :param language_code: The language code to update to.
        :raise LanguageCodeNotSupported: If the provided language code is not supported.
        """
        if (
                language_code in self.__text_message.texts_messages and
                language_code in self.__inline_keyboard.texts_buttons
        ):
            await self.state.update_data(language_code=language_code)
            self.user.language_code = language_code
            self.__text_message.language_code = self.__inline_keyboard.language_code = language_code
            return None

        raise LanguageCodeNotSupported(
            f"Language code '{language_code}' not in text message or button text"
        )

    async def open_connect_wallet_window(
            self, callbacks: Optional[ConnectWalletCallbacks] = None
    ) -> None:
        """
        Open the connect wallet window.

        :param callbacks: Callbacks to execute.
        """

        if self.tonconnect.connected:
            with suppress(WalletNotConnectedError):
                await self.tonconnect.disconnect()

        if callbacks:
            await self.connect_wallet_callbacks_storage.add(callbacks)

        task = asyncio.create_task(self.__wait_connect_wallet_task())
        self.task_storage.add(task)

        state_data = await self.state.get_data()
        wallets = await self.tonconnect.get_wallets()

        app_wallet = AppWallet.from_dict(state_data.get("app_wallet", wallets[0].to_dict()))
        universal_url = await self.tonconnect.connect(app_wallet.to_dict())

        qrcode_url = self.__qrcode_provider.create_connect_wallet_url(universal_url, app_wallet.image)
        text = hide_link(qrcode_url) + self.__text_message.get("connect_wallet").format(
            wallet_name=app_wallet.name
        )
        reply_markup = self.__inline_keyboard.connect_wallet(
            wallets, app_wallet, universal_url,
            wallet_name=app_wallet.name,
        )

        await self.state.update_data(app_wallet=app_wallet.to_dict())
        await self.send_message(text=text, reply_markup=reply_markup)
        await self.state.set_state(TcState.connect_wallet)

    async def open_send_transaction_window(
            self,
            callbacks: Optional[SendTransactionCallbacks] = None,
            transaction: Optional[Transaction] = None,
    ) -> None:
        """
        Open the send transaction window.

        :param callbacks: Callbacks to execute.
        :param transaction: The transaction details.
        """

        if transaction:
            await self.state.update_data(transaction=transaction.to_dict())

        if callbacks:
            await self.send_transaction_callbacks_storage.add(callbacks)

        task = asyncio.create_task(self.__wait_send_transaction_task())
        self.task_storage.add(task)

        text = self.__text_message.get("send_transaction").format(
            wallet_name=self.user.app_wallet.name,
        )
        universal_link = self.user.app_wallet.universal_url
        if self.user.app_wallet.app_name == "telegram-wallet":
            universal_link = universal_link + "&startattach=tonconnect"  # noqa
        reply_markup = self.__inline_keyboard.send_transaction(
            self.user.app_wallet.name, universal_link,
        )

        await self.send_message(text=text, reply_markup=reply_markup)
        await self.state.set_state(TcState.send_transaction)

    async def _connect_wallet_timeout(self) -> None:
        """
        Handle the connect wallet timeout.
        """
        text = self.__text_message.get("connect_wallet_timeout")
        reply_markup = self.__inline_keyboard.send_transaction_timeout()

        await self.send_message(text=text, reply_markup=reply_markup)
        await self.state.set_state(TcState.connect_wallet_timeout)

    async def _send_transaction_timeout(self) -> None:
        """
        Handle the send transaction timeout.
        """
        text = self.__text_message.get("send_transaction_timeout")
        reply_markup = self.__inline_keyboard.send_transaction_timeout()

        await self.send_message(text=text, reply_markup=reply_markup)
        await self.state.set_state(TcState.send_transaction_timeout)

    async def _send_transaction_rejected(self) -> None:
        """
        Handle the send transaction rejection.
        """
        text = self.__text_message.get("send_transaction_rejected")
        reply_markup = self.__inline_keyboard.send_transaction_rejected()

        await self.send_message(text=text, reply_markup=reply_markup)
        await self.state.set_state(TcState.send_transaction_rejected)

    async def send_message(
            self,
            text: str,
            reply_markup: Optional[InlineKeyboardMarkup] = None,
    ) -> Message:
        """
        Send or edit a message to the user.

        This method attempts to edit the existing message identified by the stored message ID. If editing is not
        possible (e.g., due to a message not found error), it sends a new message and deletes the previous one.

        :param text: The text content of the message.
        :param reply_markup: Optional InlineKeyboardMarkup for the message.
        :return: The edited or sent Message object.
        :raises TelegramBadRequest: If there is an issue with sending or editing the message.
        """
        state_data = await self.state.get_data()
        message_id = state_data.get("message_id", None)

        try:
            message = await self.bot.edit_message_text(
                text=text,
                chat_id=self.user.id,
                message_id=message_id,
                reply_markup=reply_markup,
            )
        except TelegramBadRequest as ex:
            if not any(e in ex.message for e in MESSAGE_EDIT_ERRORS):
                raise ex
            message = await self.bot.send_message(
                text=text,
                chat_id=self.state.key.chat_id,
                reply_markup=reply_markup,
            )
            await self._delete_previous_message()
        await self.state.update_data(message_id=message.message_id)

        return message

    async def _delete_previous_message(self) -> Union[Message, None]:
        """
        Delete the previous message.

        This method attempts to delete the previous message identified by the stored message ID. If deletion is not
        possible (e.g., due to a message not found error), it attempts to edit the previous message with a placeholder
        emoji. If editing is also not possible, it raises TelegramBadRequest with the appropriate error message.

        :return: The edited Message object or None if no previous message was found.
        :raises TelegramBadRequest: If there is an issue with deleting or editing the previous message.
        """
        state_data = await self.state.get_data()

        message_id = state_data.get("message_id", None)
        if not message_id: return  # noqa:E701

        try:
            await self.bot.delete_message(
                message_id=message_id,
                chat_id=self.user.id,
            )
        except TelegramBadRequest as ex:
            if any(e in ex.message for e in MESSAGE_DELETE_ERRORS):
                try:
                    return await self.bot.edit_message_text(
                        message_id=message_id,
                        chat_id=self.user.id,
                        text=self.__emoji,
                    )
                except TelegramBadRequest as ex:
                    if not any(e in ex.message for e in MESSAGE_EDIT_ERRORS):
                        raise ex

    async def __wait_connect_wallet_task(self) -> None:
        """
        Wait for the connect wallet task.

        This method checks the AiogramTonConnect connection status periodically for up to 3 minutes (360 iterations).
        If the connection is restored, it updates the account wallet details, executes the appropriate callbacks,
        and removes the task from the task storage. If the connection is not restored within the timeout, it
        triggers the connect wallet timeout handling.

        :raises asyncio.CancelledError: If the task is cancelled.
        :raises Exception: Any unexpected exception during the process.
        """
        try:
            for _ in range(1, 360):
                await asyncio.sleep(.5)
                await self.tonconnect.restore_connection()

                if self.tonconnect.connected:
                    state_data = await self.state.get_data()
                    app_wallet = AppWallet.from_dict(state_data.get("app_wallet"))
                    account_wallet = AccountWallet(
                        address=self.tonconnect.account.address,
                        state_init=self.tonconnect.account.wallet_state_init,
                        public_key=self.tonconnect.account.public_key,
                        chain=self.tonconnect.account.chain,
                    )

                    self.middleware_data["account_wallet"] = account_wallet
                    self.middleware_data["app_wallet"] = app_wallet
                    await self.state.update_data(account_wallet=account_wallet.to_dict())

                    callbacks = await self.connect_wallet_callbacks_storage.get()
                    await callbacks.after_callback(**self.middleware_data)
                    self.task_storage.remove()
                    return None

            await self._connect_wallet_timeout()

        except asyncio.CancelledError:
            pass
        except Exception:
            raise

        return None

    async def __wait_send_transaction_task(self) -> None:
        """
        Wait for the send transaction task.

        This method waits for the Tonconnect to send a transaction within a timeout of 5 minutes. If the transaction
        is sent successfully, it updates the user's last transaction details, executes the appropriate callbacks, and
        removes the task from the task storage. If the user rejects the transaction, it triggers the send transaction
        rejection handling. If the transaction is not sent within the timeout, it triggers the send transaction
        timeout handling.

        :raises UserRejectsError: If the user rejects the transaction.
        :raises asyncio.TimeoutError: If the transaction is not sent within the timeout.
        :raises asyncio.CancelledError: If the task is cancelled.
        :raises Exception: Any unexpected exception during the process.
        """
        try:
            data = await self.state.get_data()
            transaction = data.get("transaction")

            result = await asyncio.wait_for(
                self.tonconnect.send_transaction(transaction=transaction),
                timeout=300,
            )

            if result:
                last_transaction_boc = result.get("boc")
                self.user.last_transaction_boc = last_transaction_boc
                await self.state.update_data(last_transaction_boc=last_transaction_boc)

                callbacks = await self.send_transaction_callbacks_storage.get()
                self.middleware_data["transaction_boc"] = last_transaction_boc
                await callbacks.after_callback(**self.middleware_data)
                self.task_storage.remove()
                return None

        except UserRejectsError:
            current_state = await self.state.get_state()

            if current_state != TcState.send_transaction.state:
                return None
            await self._send_transaction_rejected()

        except asyncio.TimeoutError:
            current_state = await self.state.get_state()

            if current_state != TcState.send_transaction.state:
                return None
            await self._send_transaction_timeout()

        except asyncio.CancelledError:
            pass
        except Exception:
            raise

        return None



import base64
from abc import ABCMeta, abstractmethod


class QRCodeProviderBase(metaclass=ABCMeta):
    """
    Abstract base class for QR code providers.

    Subclasses must implement the 'create_connect_wallet_url' method.
    """

    @abstractmethod
    def create_connect_wallet_url(self, universal_url: str, wallet_image: str) -> str:
        """
        Create a URL for connecting a wallet and generate a QR code.

        :param universal_url: Universal URL for connecting the wallet.
        :param wallet_image: Image URL associated with the wallet.
        :return: URL for generating the QR code.
        """
        raise NotImplementedError


class QRCodeProvider(QRCodeProviderBase):
    """
    Default implementation of the QRCodeProviderBase.

    Uses the 'qrcode.ness.su' service to create a QR code with specific parameters.
    """

    def create_connect_wallet_url(self, universal_url: str, wallet_image: str) -> str:
        """
        Create a URL for connecting a wallet and generate a QR code.

        :param universal_url: Universal URL for connecting the wallet.
        :param wallet_image: Image URL associated with the wallet.
        :return: URL for generating the QR code.
        """
        encoded_universal_url = base64.b64encode(universal_url.encode()).decode()
        encoded_wallet_image = base64.b64encode(wallet_image.encode()).decode()
        return (
            f"https://qrcode.ness.su/create?box_size=20&border=7&image_padding=20"
            f"&data={encoded_universal_url}"
            f"&image_url={encoded_wallet_image}"
        )



from asyncio import Task
from typing import Dict

# Dictionary to store user-specific tasks
tasks: Dict[int, Task] = {}


class TaskStorage:
    """
    Storage for managing user-specific asynchronous tasks.

    This class provides methods to add, get, and remove tasks associated with a user ID.
    """

    def __init__(self, user_id: int) -> None:
        """
        Initialize the TaskStorage.

        :param user_id: User ID associated with the task storage.
        """
        self.user_id: int = user_id

    def add(self, task: Task) -> None:
        """
        Add a task to the storage.

        If a task already exists for the user, it is removed before adding the new task.

        :param task: Task to be added.
        """
        if self.user_id in tasks:
            self.remove()
        tasks[self.user_id] = task

    def get(self) -> Task:
        """
        Get the task associated with the user.

        :return: Task associated with the user, or None if not found.
        """
        return tasks.get(self.user_id)

    def remove(self) -> None:
        """
        Remove the task associated with the user.

        If the task exists and is not done, it is canceled.
        """
        task = self.get()
        if task and not task.done():
            task.cancel()
        tasks.pop(self.user_id, None)



from abc import ABCMeta, abstractmethod
from typing import Dict

from aiogram.utils.markdown import hide_link


class TextMessageBase(metaclass=ABCMeta):
    """
    Abstract base class for handling text messages in different languages.

    :param language_code: The language code for the text messages.
    """

    @property
    @abstractmethod
    def texts_messages(self) -> Dict[str, Dict[str, str]]:
        """
        Property to retrieve a dictionary of text messages for different languages.

        :return: Dictionary containing text messages for each language.
        """
        raise NotImplementedError

    def __init__(self, language_code: str) -> None:
        """
        Initialize the TextMessageBase with a specified language code.

        If the provided language code is not supported, the default language is set to "en" (English).

        :param language_code: The language code for the text messages.
        """
        if language_code not in self.texts_messages.keys():
            language_code = "en"
        self.language_code = language_code

    def get(self, code: str) -> str:
        """
        Get a specific text message based on the language code and message code.

        :param code: Code identifying the specific message.
        :return: The text message.
        """
        return self.texts_messages[self.language_code][code]


class TextMessage(TextMessageBase):
    """
    Concrete implementation of TextMessageBase providing text messages for different languages.
    """

    @property
    def texts_messages(self) -> Dict[str, Dict[str, str]]:
        get_a_wallet_link = "https://ton.org/wallets?filters[wallet_features][slug][$in]=dapp-auth&pagination[limit]=-1"
        ton_connect_banner_link = "https://telegra.ph//file/a4ddc111ff41692ad5200.jpg"

        return {
            "ru": {
                "connect_wallet": (
                    f"<a href='{get_a_wallet_link}'>Установить кошелек</a>\n\n"
                    "<b>Подключите свой {wallet_name}!</b>\n\n"
                    "Отсканируйте с помощью мобильного кошелька:"
                ),
                "connect_wallet_timeout": (
                    f"{hide_link(ton_connect_banner_link)}"
                    "<b>Предупреждение</b>\n\n"
                    "Время ожидания подключения истекло."
                ),
                "send_transaction": (
                    f"{hide_link(ton_connect_banner_link)}"
                    "<b>Транзакция</b>\n\n"
                    "Перейдите в приложение {wallet_name} и подтвердите транзакцию."
                ),
                "send_transaction_timeout": (
                    f"{hide_link(ton_connect_banner_link)}"
                    "<b>Предупреждение</b>\n\n"
                    "Время ожидания подтверждения транзакции истекло."
                ),
                "send_transaction_rejected": (
                    f"{hide_link(ton_connect_banner_link)}"
                    "<b>Предупреждение</b>\n\n"
                    "Вы отменили транзакцию!"
                ),
            },
            "en": {
                "connect_wallet": (
                    f"<a href='{get_a_wallet_link}'>Get a Wallet</a>\n\n"
                    "<b>Connect your {wallet_name}!</b>\n\n"
                    "Scan with your mobile app wallet:"
                ),
                "connect_wallet_timeout": (
                    f"{hide_link(ton_connect_banner_link)}"
                    "<b>Warning</b>\n\n"
                    "The connection timeout has expired."
                ),
                "send_transaction": (
                    f"{hide_link(ton_connect_banner_link)}"
                    f"<b>Transaction</b>\n\n"
                    "Go to the {wallet_name} app and confirm the transaction."
                ),
                "send_transaction_timeout": (
                    f"{hide_link(ton_connect_banner_link)}"
                    "<b>Warning</b>\n\n"
                    "The transaction timeout has expired."
                ),
                "send_transaction_rejected": (
                    f"{hide_link(ton_connect_banner_link)}"
                    "<b>Warning</b>\n\n"
                    "You rejected the transaction!"
                ),
            },
        }



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="aiogram-tonconnect",
    version="0.0.9",
    author="nessshon",
    description="aiogram-tonconnect seamlessly integrates TonConnect into aiogram bots, "
                "simplifying wallet connections and transactions with a ready UI. "
                "Effortless middleware for enhancing your aiogram-based bots.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(exclude=["example"]),
    python_requires='>=3.7, <3.11',
    install_requires=[
        "aiofiles~=23.1.0",
        "aiogram==3.*",
        "aiohttp~=3.8.6",
        "cachetools~=5.3.2",
        "pydantic~=2.3.0",
        "pytonconnect~=0.3.0",
        "pytoniq-core~=0.1.14",
        "redis~=5.0.1",
    ],
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    url="https://github.com/nessshon/aiogram-tonconnect/",
    project_urls={
        "Documentation": "https://nessshon.github.io/aiogram-tonconnect/",
        "Bot example": "https://t.me/aiogramTONConncetBot/",
    },
)



from typing import Optional

from pytonconnect.storage import IStorage
from redis.asyncio import Redis


class SessionStorage(IStorage):
    """
    Implementation of IStorage for session storage using Redis.

    This class provides methods to set, get, and remove items from the Redis session storage.
    """

    def __init__(
            self,
            redis: Redis,
            user_id: Optional[int] = None,
    ) -> None:
        """
        Initialize the SessionStorage.

        :param redis: Redis instance for asynchronous operations.
        :param user_id: User ID associated with the session storage.
        """
        self.redis = redis
        self.user_id = user_id

    def _get_key(self, key: str) -> str:
        """
        Generate a unique key for the session storage.

        :param key: Original key.
        :return: Unique key combining the user_id and original key.
        """
        return f"{self.user_id}{key}"

    async def set_item(self, key: str, value: str) -> None:
        """
        Set an item in the session storage.

        :param key: Key for the item.
        :param value: Value to be stored.
        """
        async with self.redis.client() as client:
            await client.set(name=self._get_key(key), value=value)

    async def get_item(self, key: str, default_value: str = None) -> str:
        """
        Get an item from the session storage.

        :param key: Key for the item.
        :param default_value: Default value if the key is not found.
        :return: Value associated with the key.
        """
        async with self.redis.client() as client:
            value = await client.get(name=self._get_key(key))
            return value.decode() if value else default_value

    async def remove_item(self, key: str) -> None:
        """
        Remove an item from the session storage.

        :param key: Key for the item to be removed.
        """
        async with self.redis.client() as client:
            await client.delete(self._get_key(key))
